#!/usr/bin/perl -w

our $VERSION='1.20130110';

=head1 NAME

foiltool.pl - view, compare, normalize, translate, rotate, and/or spline an input airfoil coordinates .dat file (or folder)

=head1 SYNOPSIS

 perl foiltool.pl  input_data_file/fodler [options]

 e.g:-

 perl foiltool.pl  coord_seligFmt/s1221.dat -x 2500 -y 1400  

     Options:
       -help	-?	brief help message
       -man		full documentation
       -x		size of the display-area create (x) (actual window will be bigger to accomdate decorations/buttons etc).
       -y		size of the display-window to create (y)
       -s		spline factor (0=4-point scheme [NDG87, DD89](default), 0.5 is the scheme promoted in [Ros04], and 1 is the uniform cubic B-spline scheme [Sab02, LR80])
			factor 0 forces the spline to pass through your supplied coords exactly.  larger numbers will "bend" inside your coods, missing many of them to stay smoothe (resultant usage might require re-normalizing)
       -l		spline subdivision level (default 2, can use 1 though 5+ if necessary (larger 'l' on sparse data=smooth output, larger 'l' on big data makes surfaces "bumpy")
       -nofix		do not scale, translate, and rotate to normalize the input
       -nospline	do NOT apply the chosen spline to the input data (after normalizing) before writing the output file
       -out file.dat	output the normalize foil data to the named file
       -format string	perl sprintf format to use (eg: "%.5f %.5f")
       -noline		don't "connect the dots" with straight green lines
       -noshow		do not plot the result (use for batch normalizing)
       -debug		print out all the data points as we manipulate them
       -sim 		compare 2 foils shape similarity (supply the folder name instead of .dat name to recursively process all foils in a folder)
       -gran		a granularity number (how precise to make the comparison) - e.g. 100 - imagine pixel width on a monitor.

 The displayed output contains:

 the input foil (as normalized) in green.
 the input data points (as normalized) as red *'s
 the output spline in blue
 a copy of the input data, trnaslated down by 20 pixels, in back (so you can easily see the input and output difference)

=head1 CREDITS

Adapted from http://www.perlmonks.org/?node_id=980572

=head1 HOW TO SANITIZE A WHOLE FOLDER OF AIRFOIL .DAT FILES

create the folder of foils - eg:-
	mkdir coord_seligFmt

create a new folder for the normalized versions:-
	mkdir norm_coord_seligFmt

run the normlaize:-
	ls coord_seligFmt/*dat | perl -ne 'chomp;$cmd="perl foiltool.pl $_ -nospline -out norm_$_ -noshow"; print "$cmd\n"; print `$cmd`'

create another new folder for the XFoil-fixed (via PANE) versions:-
	mkdir pane_norm_coord_seligFmt

run xfoil to fix them all:-
	ls norm_coord_seligFmt/*.dat |perl -ne 'chomp;open(O,">in");print O "\n\nNORM\n\nPLOP\nG\n\n\nLOAD $_\nPANE\nSAVE pane_$_\nQUIT\n\n";close(O);$x=`xfoil <in`;' 

Your files now live in pane_norm_coord_seligFmt.  

View a new one:
	perl foiltool.pl pane_norm_coord_seligFmt/e49.dat  -x 2500 -y 1400 -nofix

View the original of it:
	perl foiltool.pl coord_seligFmt/e49.dat  -x 2500 -y 1400 -nofix

=head1 DESCRIPTION

Uses Gtk2 windows to plot airfoil datapoints.

=head1 NOTES

see page 40: Wing_Design-19960015887_1996034550.pdf (switch to crescent @ r=0.6)

=head1 SUBROUTINES

=cut

######################################################################

use strict;
use warnings;		# same as -w switch above

use Glib qw/TRUE FALSE/;
use Gtk2; # -init;
my $TRUE=TRUE; my $FALSE=FALSE;


use Getopt::Long;	# Commandline argument parsing
use Pod::Usage;		# Inbuilt documentation helper
use Math::Trig;

use constant PI    => 4 * atan2(1, 1);
my $slash='/'; $slash="\\" if($^O =~/Win32/i);


my $area; 
my %allocated_colors;
my ( $x0, $y0, $x1, $y1, $width, ) = ( 0, 0, 0, 0 );
my $msg='';
my @points;
my %foils;


my %arg;		# We store copies of commandline arguments in here

# Parse options, and print help/manual if requested.
&GetOptions('help|?' => \$arg{'help'},			# breif instructions
	    'man' => \$arg{'man'},			# complete manual
	    'x=i'=>\$arg{'x'},				# screen X
	    'y=i'=>\$arg{'y'},				# screen Y
	    'l=i'=>\$arg{'l'},				# screen Y
	    's=s'=>\$arg{'s'},				# spline type
	    'out=s'=>\$arg{'out'},			# file to write normalized data into
	    'sim=s'=>\$arg{'sim'},			# similarity routine
	    'gran=i'=>\$arg{'gran'},			# similarity granularity
	    'format=s'=>\$arg{'format'},		# output number formatting (not a good idea to use - truncates precision)
	    'nofix'=>\$arg{'nofix'},			# do not scale, translate, and rotate to normalize the input
	    'debug'=>\$arg{'debug'},			
	    'noshow'=>\$arg{'noshow'},			
	    'nospline'=>\$arg{'nospline'},			
	    'tst=s'=>\$arg{'tst'},			
	    'noline'=>\$arg{'noline'},			# don't "connect the dots" with straight green lines
	   ) or &pod2usage(2); 
no warnings;
	   &pod2usage(1) if ($arg{'help'});
	   &pod2usage(-exitstatus => 0, -verbose => 2) if $arg{'man'};
use warnings;

my $subdivision_level = 2; $subdivision_level =$arg{'l'} if($arg{'l'});

# Decide on the size of the window to open
my $xsize             = 640; $xsize=$arg{'x'} if($arg{'x'});
my $ysize             = 480; $ysize=$arg{'y'} if($arg{'y'});
my $winx=$xsize+30;	# 640 yeilds 648 (displayable X=610)
my $winy=$ysize+81;	# 480 yeilds 509 (displayable Y=399)


$arg{'s'}=0 unless(defined($arg{'s'}));
my ($xo,$yo,$xm,$ym)=($xsize/2,$ysize/2,$xsize/1.02,-$xsize/1.02);	# ym scaled by $xsize to retain aspect ratio

if (defined($arg{'tst'})) {
  my $ret=&Ypoint(split(/,/,$arg{'tst'}));
  print "Ypoint(x,maxx,maxy,minx,miny)=$ret\n";
  exit(0);
}


if(!defined $ARGV[0]) {
  warn "Must supply input foil or folder";
  &pod2usage(1) if ($arg{'help'});
} elsif(-d  $ARGV[0]) {
  &ProcDir( $ARGV[0] ,\&ProcFoil );
} else {
  &ProcFoil( $ARGV[0] );
}

print "# Run complete.\n";
exit(0);



sub HashFoil {
  my($ffn)=@_;
  @{$foils{$ffn}}=&LoadFoil($ffn);
}



sub ProcFoil {
  my($hd)=@_; # Input foil name

  @points=&LoadFoil($hd,$arg{'out'});

  if($arg{'sim'}) {
    if(-d $arg{'sim'}) {
      &ProcDir($arg{'sim'}, \&HashFoil)
    } else {
      @{$foils{$arg{'sim'}}}=&LoadFoil($arg{'sim'});
    }
#    my @points2=&LoadFoil($arg{'sim'});
    foreach my $ffn (sort keys %foils) {
      my $simi=&Sim(\@points,$foils{$ffn},$arg{'gran'});
      print "$simi\t$hd\t$ffn\n";	# Output the data in a sortable way
    }

  }


  return if($arg{'noshow'});

print "$msg\n";

Gtk2->init;

my $area;

my %allocated_colors;
my ( $x0, $y0, $x1, $y1, $width, ) = ( 0, 0, 0, 0 );

my $window = new Gtk2::Window("toplevel");
$window->signal_connect( "delete_event", sub { Gtk2->main_quit; } );
$window->set_border_width(10);
$window->set_size_request( $winx, $winy );
$window->set_position('center');

my $vbox = Gtk2::VBox->new( 0, 0 );
$window->add($vbox);
$vbox->set_border_width(2);

my $hbox = Gtk2::HBox->new( 0, 0 );
$vbox->pack_start( $hbox, 1, 1, 0 );
$hbox->set_size_request( 320, 240 );
$hbox->set_border_width(2);

my $hbox1 = Gtk2::HBox->new( 0, 0 );
$vbox->pack_start( $hbox1, 0, 0, 0 );
$hbox1->set_border_width(2);

my $button1 = Gtk2::Button->new('Generate');
$hbox1->pack_start( $button1, $FALSE, $FALSE, 2 );
$button1->signal_connect( clicked => sub { start_drawing($area) } );

my $button2 = Gtk2::Button->new('Exit');
$hbox1->pack_start( $button2, $FALSE, $FALSE, 2 );
$button2->signal_connect( clicked => sub { exit; } );

my $scwin = Gtk2::ScrolledWindow->new();
my $ha1   = $scwin->get_hadjustment;
$scwin->set_policy( 'never', 'never' );
#$scwin->set_policy( 'always', 'never' );

my $vp = Gtk2::Viewport->new( undef, undef );
$scwin->add($vp);
$hbox->pack_start( $scwin, 1, 1, 0 );

$area = new Gtk2::DrawingArea;
$area->size( $xsize, $ysize );
$vp->add($area);

$area->set_events(
	[
		qw/exposure-mask
		  leave-notify-mask
		  button-press-mask
		  pointer-motion-mask
		  pointer-motion-hint-mask/
	]
);

$area->signal_connect( button_press_event => \&button_press_event );



$window->show_all;

&draw_line( $scwin, [@points], 'green' );



Gtk2->main;

} # ProcFoil


sub get_color {
	my ( $colormap, $name ) = @_;
	my $ret;

	if ( $ret = $allocated_colors{$name} ) {
		return $ret;
	}

	my $color = Gtk2::Gdk::Color->parse($name);
	$colormap->alloc_color( $color, $TRUE, $TRUE );

	$allocated_colors{$name} = $color;

	return $color;
}

sub draw_line {
	my ( $widget, $line, $color ) = @_;

	my $colormap = $widget->window->get_colormap;

	my $gc = $widget->{gc} || new Gtk2::Gdk::GC $widget->window;
	$gc->set_foreground( get_color( $colormap, $color ) );

	$widget->window->draw_lines( $gc, @$line );
	$gc->set_foreground( get_color( $colormap, 'black' ) );
	$widget->window->draw_points( $gc, @$line );
}


# adds square of differences in foil shape, considering only slices of Y (might have been more accurate to see how far a tangent on foil 1 needs to go to intersect foil 2...)
sub Sim {
  my($pp1,$pp2,$gran)=@_;

# warn @{$pp1};
# warn @{$pp2};
# die 1;


  $gran=100 unless(defined $gran);

  my($px1,$py1)=&points_to_xy(@{$pp1});
  my($px2,$py2)=&points_to_xy(@{$pp2});
  my $simi=0; # Similarity accumulator
  my $x=0;
  for(my $i=0;$x<=1;$i+=(1/$gran)) {
    
    foreach my $topbot (-1,1) { # For the top surface, followed by the undersurface...
      my($y1)=&FindY($x,$topbot,$px1,$py1);	# Find the largest (top) Y that's at X-axis point $x on foil 1
      my($y2)=&FindY($x,$topbot,$px2,$py2);	# and foil 2
      #print "$topbot,$x,$y1,$y2\n";		# output excel .csv to observe the weighting
      $simi+=($y1-$y2)**2;			# Add the square of the differences
    } # top/bot
    my $step=2*(((1/$gran)+$i)**2)*(1/$gran);	# Weights the result to the leading edge (un-commend the above "print" and open resulting .csv in excel, & chart to see)
    $step=(1/$gran) if($step>(1/$gran));
    $x+=$step;
  } # $i (gran)

  return $simi;
} # Sim




sub FindY {	# Return a Y for a supplied X
  my ($x,$topbot,$px,$py)=@_;
  my($minx,$maxx,$miny,$maxy,$y);
  my $ret;

  # Search for either $x, or $x-1 and $x+1, and grab the y(s) appropriate.
  for(my $i=0;$i<$#{$px};$i++) {
    my $chkx=$px->[$i];

    # Top
#print "    if(( $topbot<0 )&&(($chkx==$x)||( (($chkx>$x)  && ($px->[$i+1]<$x))   ) )) \n";


    if(( $topbot<0 )&&(($chkx==$x)||( (($chkx>$x)  && ($px->[$i+1]<$x))   ) )) {
      return $py->[$i] if ($chkx==$x);
      $maxx=$px->[$i];
      $maxy=$py->[$i];
      $minx=$px->[$i+1];
      $miny=$py->[$i+1];
      $ret= &Ypoint($x,$maxx,$maxy,$minx,$miny);	# return the 1st one we find
#print "FindY($x,$topbot,$px,$py)=$ret\n";
      return $ret;
    }

    # Bottom
    if(( $topbot>0 )&&(($chkx==$x)||( (($chkx<$x)  && ($px->[$i+1]>$x))   ) )) {
      if ($chkx==$x) {
        $ret=$py->[$i]
      } else {
        $maxx=$px->[$i+1];
        $maxy=$py->[$i+1];
        $minx=$px->[$i];
        $miny=$py->[$i];
        $ret=&Ypoint($x,$maxx,$maxy,$minx,$miny); # use the last one we find (foils are drawn from TE to LE via top 1st)
      }
    }
  }
  
  unless(defined $ret) {
    warn "Could not find Y for X=$x (surface $topbot)\n";
    # die if($x==0);
    return 9;
  }
  return $ret;
}

sub Ypoint {
  my ($x,$maxx,$maxy,$minx,$miny)=@_;
  my $scalex=$maxx-$minx;
  my $scaley=$maxy-$miny;
  my $proportion=$maxx-$x;
  
  return $maxy-($proportion/$scalex)*$scaley;
} # Ypoint


sub LoadFoil {
  my($hd,$outfn)=@_;
  my @points;


  open(IN,'<',$hd) or die "Cannot read input airfoil data file '$hd': $!";

  my @foil=<IN>; close(IN);


  #ignore all 1st lines now unless($foil[0]=~/^\s*[\d\-]/){ # ignore 1st line if it doesn't begin with a number
  $hd=$foil[0];	# copy heading for output file
  $foil[0]='';
  # }


  my(@x,@y); my $warned=0;
  my($minx,$maxx,$topy,$boty,$mini);
  my $i=0;
  # Read file
  foreach (@foil) {
    next if (/^$/);
    # next if (/\.\./);
    $_=~s/(\d)([\+\-])/$1 $2/g;	# some files have no spaces - eg -0.145+0.245
    $_=~s/^\s+//;
    my($x,$y)=split(/\s+/); $x=~s/[\(\)]//g; $y=~s/[\(\)]//g; $y=0 if($y=~/\.\./);
    $x+=0; $y+=0; # convert to numeric
    # warn "file was not normalized: fixed x=$x" if((($x<0)||($x>1)) &&(!$warned++));
    if((!defined $minx)||($x<$minx)) {
      $minx=$x;$mini=$i;
    }
    $maxx=$x if((!defined $maxx)||($x>$maxx));
    $topy=$y if(!defined $topy);	# first y
    $boty=$y;				# last y
    push @x,$x; push @y,$y;
    print "($x,$y)\n" if($arg{'debug'});
    $i++;
  }

  if(!$arg{'nofix'}) { # SKIPNORM
  
    print "========== minx=$minx maxx=$maxx topy=$topy boty=$boty mini=$mini y at mini=$y[$mini]\n" if($arg{'debug'});
  
    # Normalize (scale & translate) the data so that x coordinate ranges from 0 to 1, and trailling edge Y is on (or centered around) 0 (nb: may need a re-do after rotate)
    my $scale=$maxx-$minx;
    my $yt=($topy+$boty)/2;
    my $xt=$minx;
    if(($minx!=0)||($maxx!=1)||(($topy+$boty)!=0)) {
      $msg.="scaled*" . (1/$scale) . " translated($xt,$yt) ";
      ($minx,$maxx,$topy,$boty,$mini)=();
      for(my $i=0;$i<=$#x;$i++) {
        $x[$i]=($x[$i]-$xt)/$scale;	# Translate X and scale
        $y[$i]=$y[$i]/$scale-$yt;		# scale and translate Y
        print "=>($x[$i],$y[$i])\n" if($arg{'debug'});
  
        if((!defined $minx)||($x[$i]<$minx)) {
          $minx=$x[$i]; $mini=$i;
        }
        $maxx=$x[$i] if((!defined $maxx)||($x[$i]>$maxx));
        $topy=$y[$i] if(!defined $topy);        # first y
        $boty=$y[$i];                           # last y
  
      }
      print "========== (scale was $scale yt=$yt new:minx=$minx maxx=$maxx topy=$topy boty=$boty mini=$mini y at mini=$y[$mini]\n" if($arg{'debug'});
    }
  
  #REROTATE:
  
    if(abs($y[$mini])>0.00001) {
      # rotate foil so that leading edge min X coord always has Y=0
      my $rot=asin($y[$mini])/(PI/180);
      &rotatexy($rot,1,0,\@x,\@y); # probably only works for small initial mismatches
      $msg.="rotated x $rot degrees ";
      print "========== rotated by $rot degrees\n" if($arg{'debug'});
  
      ($minx,$maxx,$topy,$boty,$mini)=();
      for(my $i=0;$i<=$#x;$i++) {
        if((!defined $minx)||($x[$i]<$minx)) {
          $minx=$x[$i]; $mini=$i;
        }
        $maxx=$x[$i] if((!defined $maxx)||($x[$i]>$maxx));
        $topy=$y[$i] if(!defined $topy);        # first y
        $boty=$y[$i];                           # last y
      }
      print "========== (new post-rot minx=$minx maxx=$maxx topy=$topy boty=$boty mini=$mini y at mini=$y[$mini]\n" if($arg{'debug'});
    }
  
    # re-Normalize (scale & translate) in case the above rotation was dramatic
    $scale=$maxx-$minx;
    $yt=($topy+$boty)/2;
    $xt=$minx;
    if(($minx!=0)||($maxx!=1)||(($topy+$boty)!=0)) {
      $msg.="then(again) scaled*" . (1/$scale) . " translated($xt,$yt) ";
      ($minx,$maxx,$topy,$boty,$mini)=();
      for(my $i=0;$i<=$#x;$i++) {
        $x[$i]=($x[$i]-$xt)/$scale;	# Translate X and scale
        $y[$i]=$y[$i]/$scale-$yt;		# scale and translate Y
        print "=>($x[$i],$y[$i])\n" if($arg{'debug'});
  
        if((!defined $minx)||($x[$i]<$minx)) {
          $minx=$x[$i]; $mini=$i;
        }
        $maxx=$x[$i] if((!defined $maxx)||($x[$i]>$maxx));
        $topy=$y[$i] if(!defined $topy);        # first y
        $boty=$y[$i];                           # last y
  
      }
      print "========== (post rot scale was $scale yt=$yt new:minx=$minx maxx=$maxx topy=$topy boty=$boty mini=$mini y at mini($i)=$y[$mini]\n" if($arg{'debug'});
    }
  
    # could, if needed, loop back to REROTATE if we need better rotation precision for badly aligned input foil data
  
    my @savepoints=&xy_to_points(\@x,\@y);
  
    # Spline our data unless not wanted
    if(!$arg{'nospline'}) {
      my $sl=$subdivision_level;
      $sl++ if(($#x<=50) && (!defined($arg{'l'})));	# low numbers of data points want a smoothe spline
      $msg.="splined via factor $arg{'s'}, level $sl from " . (1+$#x) . " coords to ";
      my @points=&xy_to_points(\@x,\@y); 	# foreach (@points) {$_+=3;$_*=10}; #warn "before @points";
      @points=&DoSplineReal($sl,@points); 	# foreach (@points) {$_/=10;$_-=3};
      my($px,$py)=&points_to_xy(@points);
      @x=@{$px}; @y=@{$py};
      $msg.=(1+$#x) . " coords ";
      #warn "after @points";
    }
    
    
    if(defined $outfn) {
      chop $hd while((substr($hd,-1) eq "\015")||(substr($hd,-1) eq "\012"));
      open(OUT,'>',$outfn) or die "Cannot open output file '$outfn': $!";
      chop($msg);
      print OUT "$hd (!! Normalized by " . getpwuid( $< ) . " with $0 v$VERSION on ${\GetNow()}: $msg)\n";
      for(my $i=0;$i<=$#x;$i++) {
        if($arg{'format'}) {
          print OUT sprintf($arg{'format'},$x[$i],$y[$i]) . "\n";
        } else {
          print OUT "$x[$i]\t$y[$i]\n";
        }
      }
      close(OUT);
    }
    
    if(1) { # restore the original points - so the screen shows a plot of what we already wrote to file
      my($px,$py)=&points_to_xy(@savepoints);
      @x=@{$px}; @y=@{$py};
    }
    
  
  } # SKIPNORM:
  
  # adjust point data to fit our display size
    for(my $i=0;$i<=$#x;$i++) {
  if(!$arg{'sim'}) {
      $x[$i]=($x[$i]-0.5)*$xm+$xo;	# center the foil at screen middle
      $y[$i]=$y[$i]*$ym+$yo;
  }
      print "=>($x[$i],$y[$i])\n" if($arg{'debug'});
      push @points, $x[$i];
      push @points, $y[$i];
    }

  print "==========\n" if($arg{'debug'});

  return @points;
} # LoadFoil








sub xy_to_points {
  my($px,$py)=@_;
  my @points;
  for(my $i=0;$i<=$#{$px};$i++) {
    push @points,$px->[$i];
    push @points,$py->[$i];
  }
  @points;
} # xy_to_points


sub points_to_xy {
  my @points=@_;
  my @xy; my $t=''; foreach (@points) {if($t eq '') {$t.=$_} else {push @xy,"$t,$_";$t=''}}
  my(@x,@y);
	foreach (@xy) {
	  my($x,$y)=split(/,/);
	  push @x,$x;
	  push @y,$y;
	}
  return (\@x,\@y);
# test worked: my($ptx,$pty)=&points_to_xy(1,2,3,4); warn "$ptx,$pty"; my @tst=&xy_to_points($ptx,$pty); warn "@tst";
} # points_to_xy


sub yoffset {
  my ($amount,@points)=@_;
  my($ptx,$pty)=&points_to_xy(@points);
  foreach (@{$pty}){$_+=$amount};
  return &xy_to_points($ptx,$pty);
} # yoffset

# rotates the data pointed to by x and y
sub rotatexy {
  my ($a,$aboutx,$abouty,$px,$py)=@_; $a=$a*(PI/180);

  for(my $i=0;$i<=$#{$px};$i++) {

    my $x=$px->[$i]-$aboutx;
    my $y=$py->[$i]-$abouty;

print "$aboutx($x,$y)=>(" if($arg{'debug'});

    $px->[$i]=$x*cos($a)+$y*sin($a)+$aboutx;
    $py->[$i]=$y*cos($a)+$x*sin($a)+$abouty;

print $px->[$i] . "," . $py->[$i] . ")\n" if($arg{'debug'});

  }
} # rotate


sub rotate {
  my ($a,$aboutx,$abouty,@points)=@_;
  my($px,$py)=&points_to_xy(@points);
  &rotatexy($a,$aboutx,$abouty,$px,$py);
  return &xy_to_points($px,$py);
}


sub rotatep {
  my ($a,$aboutx,$abouty,@points)=@_; $a=$a*(PI/180);
  my($px,$py)=&points_to_xy(@points);
  for(my $i=0;$i<=$#{$px};$i++) {


    my $x=$px->[$i]-$aboutx;
    my $y=$py->[$i]-$abouty;

print "$aboutx($x,$y)=>(" if($arg{'debug'});

    $px->[$i]=$x*cos($a)+$y*sin($a)+$aboutx;
    $py->[$i]=$y*cos($a)+$x*sin($a)+$abouty;

print $px->[$i] . "," . $py->[$i] . ")\n" if($arg{'debug'});


  }
  return &xy_to_points($px,$py);
} # rotate

sub start_drawing {
	my $area = shift;

&draw_line( $area, [0,$ysize/2,$xsize-1,$ysize/2], 'yellow' );	# show grid lines
&draw_line( $area, [$xsize/2,0,$xsize/2,$ysize-1], 'yellow' );


	&draw_line( $area, [@points], 'green' ) unless($arg{'noline'});	# Original data we're about to overwrite with our spline
	&draw_line( $area, [&yoffset(20,@points)], 'black' ) unless($arg{'noline'});	# same - offset a bit so we can see it easily

# below - was from testing rotation:-
# my @tmppoints=&yoffset(40,@points); 	&draw_line( $area, [&rotate(10,$xo,$yo,@tmppoints)], 'cyan' ) unless($arg{'noline'});

	my @xy; my $t=''; foreach (@points) {if($t eq '') {$t.=$_} else {push @xy,"$t,$_";$t=''}}
	my $lx=int(($ysize/200)+0.5); # length of the "+" marker lines
	foreach (@xy) {
	  my($x,$y)=split(/,/);
	  &draw_line( $area, [$x,$y-$lx,$x,$y+$lx], 'red' );	# put a red X on every datapoint
	  &draw_line( $area, [$x-$lx,$y,$x+$lx,$y], 'red' );
	  &draw_line( $area, [$x-$lx,$y-$lx,$x+$lx,$y+$lx], 'red' );	# put a red * on every datapoint
	  &draw_line( $area, [$x+$lx,$y-$lx,$x-$lx,$y+$lx], 'red' );
	}

#&draw_line( $area, [2,2,638,2,638,478,2,478,2,2], 'red' );
# &draw_line( $area, [0,0,$xsize-1,0,$xsize-1,$ysize-1,0,$ysize-1,0,0], 'red' );

&draw_line( $area, [$xsize/2-1,$ysize/2,$xsize/2+1,$ysize/2], 'red' );


  @points=&DoSpline(@points);	# Calculate a new spline


	&draw_line( $area, [@points], 'blue' );	# show it

#	print "@points";
}


sub DoSplineReal {
 my($subdivision_level, @points)=@_;
	my $a;
	my $b;
	my $k = 0;
	while ( $k++ < $subdivision_level ) {
		my $j = 0;
		my @tmp;
		while ( $j < $#points ) {
			$a = $b = 0;
			if(defined $arg{'s'}) {$a = $b = $arg{'s'}}
			if ( $j == 0 ) {			# Anchor start of output line to the start point

				push( @tmp, $points[$j] );
				push( @tmp, $points[ $j + 1 ] );
#print "a $j\t(" . $points[$j] . "," . $points[$j+1] . ")\n";
			}
			elsif ( $j + 3 <= $#points ) {		# 
				my ( $pt1, $pt2 );

				$pt1 =
				  ( $a * $points[ $j - 2 ] +
					  ( 8 - 2 * $a ) * $points[$j] +
					  $a * $points[ $j + 2 ] ) / 8;
				$pt2 =
				  ( $a * $points[ $j - 1 ] +
					  ( 8 - 2 * $a ) * $points[ $j + 1 ] +
					  $a * $points[ $j + 3 ] ) / 8;
				push( @tmp, $pt1 );
				push( @tmp, $pt2 );

#print "b $j\t($pt1,$pt2)\n";

			}
			if ( $j + 5 <= $#points && $j > 0) {
				my ( $pt1, $pt2 );

				#push( @tmp, $points[ $j + 2 ] );
				#push( @tmp, $points[ $j + 3 ] );

				$pt1 =
				  ( ( $b - 1 ) * $points[$j -2] +
					  ( 9 - $b ) * $points[ $j ] +
					  ( 9 - $b ) * $points[ $j + 2 ] +
					  ( $b - 1 ) * $points[ $j + 4 ] ) / 16;
				$pt2 =
				  ( ( $b - 1 ) * $points[ $j -1 ] +
					  ( 9 - $b ) * $points[ $j + 1 ] +
					  ( 9 - $b ) * $points[ $j + 3 ] +
					  ( $b - 1 ) * $points[ $j + 5 ] ) / 16;
				push( @tmp, $pt1 );
				push( @tmp, $pt2 );
#print "c $j\t($pt1,$pt2)\n";
			}
			$j += 2;
		}
		push( @tmp, $points[ $#points - 1 ] );
		push( @tmp, $points[$#points] );
#print "d $j\t(" . $points[ $#points - 1 ] . ",$points[$#points])\n";
		@points = ();
		@points = @tmp;
		#print "@points\n";
	}
  return @points;
} # DoSplineReal

sub DoSpline {
 my @points=@_;
	my $a;
	my $b;
	my $k = 0;
	while ( $k++ < $subdivision_level ) {
		my $j = 0;
		my @tmp;
		while ( $j < $#points ) {
			$a = $b = 0;
			if(defined $arg{'s'}) {$a = $b = $arg{'s'}}
			if ( $j == 0 ) {			# Anchor start of output line to the start point

				push( @tmp, $points[$j] );
				push( @tmp, $points[ $j + 1 ] );
#print "a $j\t(" . $points[$j] . "," . $points[$j+1] . ")\n";
			}
			elsif ( $j + 3 <= $#points ) {		# 
				my ( $pt1, $pt2 );

				$pt1 =
				  ( $a * $points[ $j - 2 ] +
					  ( 8 - 2 * $a ) * $points[$j] +
					  $a * $points[ $j + 2 ] ) / 8;
				$pt2 =
				  ( $a * $points[ $j - 1 ] +
					  ( 8 - 2 * $a ) * $points[ $j + 1 ] +
					  $a * $points[ $j + 3 ] ) / 8;
				push( @tmp, int($pt1) );
				push( @tmp, int($pt2) );

#print "b $j\t($pt1,$pt2)\n";

			}
			if ( $j + 5 <= $#points && $j > 0) {
				my ( $pt1, $pt2 );

				#push( @tmp, $points[ $j + 2 ] );
				#push( @tmp, $points[ $j + 3 ] );

				$pt1 =
				  ( ( $b - 1 ) * $points[$j -2] +
					  ( 9 - $b ) * $points[ $j ] +
					  ( 9 - $b ) * $points[ $j + 2 ] +
					  ( $b - 1 ) * $points[ $j + 4 ] ) / 16;
				$pt2 =
				  ( ( $b - 1 ) * $points[ $j -1 ] +
					  ( 9 - $b ) * $points[ $j + 1 ] +
					  ( 9 - $b ) * $points[ $j + 3 ] +
					  ( $b - 1 ) * $points[ $j + 5 ] ) / 16;
				push( @tmp, int($pt1) );
				push( @tmp, int($pt2) );
#print "c $j\t($pt1,$pt2)\n";
			}
			$j += 2;
		}
		push( @tmp, $points[ $#points - 1 ] );
		push( @tmp, $points[$#points] );
#print "d $j\t(" . $points[ $#points - 1 ] . ",$points[$#points])\n";
		@points = ();
		@points = @tmp;
		#print "@points\n";
	}
  return @points;
} # DoSpline




sub ProcDir {
  my($dir,$func)=@_;
  $dir.=$slash unless(substr($dir,-1) eq $slash);
  my(@subdir,$THISDIR);
  if(opendir($THISDIR,$dir)) {

    while(my $file=readdir $THISDIR) {
      next if(($file eq '.')||($file eq '..'));
      my $fn="$dir$file";
      #my @stat=lstat($fn);	# stats the sym, not the target # 2=type+mode, 6=rdev, 7=size, 9=mtime
      my @stat=stat($fn);	# stats the target # 2=type+mode, 6=rdev, 7=size, 9=mtime
      my($mode)=$stat[2]>>12;	# 8=regular, 10=symlink, 4=dir

      push @subdir,$fn if(($mode==4)); # &&(!-l $fn));	# recurse

      &{$func}($fn) if((-f $fn)||(-l $fn));	# files & symlinks, not anything else

    } # readdir
  } else {
      warn "can't open '$dir': $! ";
  }
  close($THISDIR);
  foreach my $sub(@subdir) {
    &ProcDir($sub,$func);
  }
} # ProcDir






sub GetNow {
  my @monthname=('Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec');
  my @weekname=('Sun','Mon','Tue','Wed','Thu','Fri','Sat');
  my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=gmtime();
  return sprintf("%s, %02d %s %04d %02d:%02d:%02d GMT",$weekname[$wday],$mday,$monthname[$mon],$year+1900,$hour,$min,$sec); 
}



=head1 AUTHOR

C. N. Drake, E<lt>christopher@pobox.comE<gt>

=head1 WARRANTY and DISCLAIMER

Important. There is no warranty (see License below).

USE AT OWN RISK!

=head1 COPYRIGHT AND LICENSE

This is distrubed under the GNU Public License:-
http://www.gnu.org/copyleft/gpl.html

Copyright (C) 2013 by C. N. Drake.

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.3 or,
at your option, any later version of Perl 5 you may have available.

=cut
